/******************************************************************************
* LockState v1.1
* Copyright © SFR 2017, 2019
* GNU GPL v3 applies (http://www.gnu.org/licenses/gpl.html)
* 
* SVG status icons © vovchik 2017
* 
* Disclaimer: my 2nd C/GTK+ app, slightly better than previous one,
* but still quite messy and I'm sure many things could've be done better...
******************************************************************************/

/* Build with:
	gcc lockstate.c -o lockstate -O3 -s `pkg-config --cflags --libs gtk+-2.0` -lX11 -lXtst
*/

/* ========================================================================= */

/* Core */
#include <X11/Xlib.h>
#include <X11/extensions/XTest.h>
#include <X11/keysym.h>
#include <glib.h>
#include <glib/gprintf.h>
#include <glib-object.h>
#include <gdk-pixbuf/gdk-pixbuf.h>
#include <gtk/gtk.h>

/* Gettext */
#include <libintl.h>
#include <locale.h>
#define _(STRING) gettext(STRING)

/* ------------------------------------------------------------------------- */

#define GDK2HEX6(dest, src) \
		g_sprintf ((dest), "#%02x%02x%02x", (src).red >> 8, (src).green >> 8, (src).blue >> 8)

/* ------------------------------------------------------------------------- */

#define APPNAME			"LockState v1.1"
#define CREDITS			"Copyright © SFR 2017, 2019 (GNU GPL v3)\n" \
						"SVG status icons created by vovchik (GNU GPL v2)"
#define WEBSITE			"http://murga-linux.com/puppy/viewtopic.php?t=110710"

#define CONF_FILE		"lockstate.conf"

#define SHOW_CAPS		TRUE
#define SHOW_NUM		TRUE
#define SHOW_SCROLL		FALSE

#define REV_ORDER		FALSE

#define FG_ON		 	"#ff0000"
#define FG_OFF		 	"#ff0000"
#define BG_ON_BEG	 	"#f0f0f0"
#define BG_ON_END	 	"#ffffff"
#define BG_OFF_BEG	 	"#888888"
#define BG_OFF_END	 	"#ffffff"

#define SYM_CAPS_ON		"A"
#define SYM_CAPS_OFF	"a"
#define SYM_NUM_ON		"1"
#define SYM_NUM_OFF		"0"
#define SYM_SCROLL_ON	"↓"
#define SYM_SCROLL_OFF	"⇣"

#define ICON_TEMPLATE \
"<svg height='32' width='32' viewBox='0 0 16 16'> \
	<radialGradient id='rg' fx='50%%' fy='50%%' cx='50%%' cy='50%%' r='50%%'> \
		<stop offset='0%%'   stop-color='%s' stop-opacity='1'/> \
		<stop offset='100%%' stop-color='%s' stop-opacity='1' /> \
	</radialGradient> \
	<rect width='14' height='14' x='1' y='1' rx='2' ry='3' fill='999999'/> \
	<rect width='14' height='13' x='1' y='1' rx='2' ry='3' fill='url(#rg)'/>  \
	<g font-size='11' font-style='normal' font-variant='normal' \
		font-weight='normal' font-stretch='normal' text-align='center' \
		line-height='125%%' letter-spacing='0' word-spacing='0' \
		text-anchor='middle' fill='%s' fill-opacity='1' \
		stroke='none' font-family='sans' dominant-baseline='central'> \
		<text x='50%%' dy='11'>%s</text> \
	</g> \
</svg>"

/* ========================================================================= */	

/* Defaults */
gboolean show_caps		= SHOW_CAPS;
gboolean show_num		= SHOW_NUM;
gboolean show_scroll	= SHOW_SCROLL;

gboolean rev_order		= REV_ORDER;

/* 8 = #rrggbb (7) + NUL (1) */
gchar fg_on[8]			= FG_ON;
gchar fg_off[8]			= FG_OFF;
gchar bg_on_beg[8]		= BG_ON_BEG;
gchar bg_on_end[8]		= BG_ON_END;
gchar bg_off_beg[8]		= BG_OFF_BEG;
gchar bg_off_end[8]		= BG_OFF_END;

/* 8, because a Unicode char may take 6 bytes + NUL + 1 byte for nice padding */
gchar sym_caps_on[8]	= SYM_CAPS_ON;
gchar sym_caps_off[8]	= SYM_CAPS_OFF;
gchar sym_num_on[8]		= SYM_NUM_ON;
gchar sym_num_off[8]	= SYM_NUM_OFF;
gchar sym_scroll_on[8]	= SYM_SCROLL_ON;
gchar sym_scroll_off[8]	= SYM_SCROLL_OFF;

/* Status icons' pixbufs */
typedef union {
	struct {
		GdkPixbuf *caps_on;	
		GdkPixbuf *caps_off;
		GdkPixbuf *num_on;
		GdkPixbuf *num_off;
		GdkPixbuf *scroll_on;
		GdkPixbuf *scroll_off;
	};
	GdkPixbuf *array[6];	/* to access members also using index */
	
} Pixbuf;

Pixbuf *pixbuf = NULL;

/* Tray icons */
GtkStatusIcon	*icon_caps;
GtkStatusIcon	*icon_num;
GtkStatusIcon	*icon_scroll;

gboolean refresh_request = FALSE;
	
/* ------------------------------------------------------------------------- */

void check_error (GError **error) {

	if (*error) g_printerr ("Error: %s\n\n", (*error)->message);
	g_clear_error (error);

}

/* ------------------------------------------------------------------------- */

void set_icons (void) {

	guint				i, j, s;
	gchar				*template;
	GdkPixbufLoader		*loader;
	gsize				len;
	GError				*error = NULL;
	
	gchar sym[8];
	gchar *src[]	= { "&",		"<",		">"		};
	gchar *rep[]	= { "&amp;",	"&lt;",		"&gt;"	};

	gchar *symbol[3][2] = {
		{ sym_caps_on,		sym_caps_off },
		{ sym_num_on,		sym_num_off },
		{ sym_scroll_on,	sym_scroll_off }
	};
	
	gchar *color_fg[]		= { fg_on,		fg_off };
	gchar *color_bg_beg[]	= { bg_on_beg,	bg_off_beg };
	gchar *color_bg_end[]	= { bg_on_end,	bg_off_end };

	Pixbuf *pixbuf_old = pixbuf;
	pixbuf = g_malloc(sizeof(*pixbuf));

	/* Create pixbufs */
	for (i = 0; i < 3; ++i) {
		for (j = 0; j < 2; ++j) {
			/* replace & < > to html entities */
			g_strlcpy (sym, symbol[i][j], -1);
			for (s = 0; s < sizeof (src) / sizeof (*src); ++s) {
				if (g_strcmp0 (sym, src[s]) == 0)
					g_strlcpy (sym, rep[s], -1);
			}

			template = g_strdup_printf (ICON_TEMPLATE, color_bg_end[j], color_bg_beg[j], color_fg[j], sym);
			/* calculate length (I don't want to incl. <string.h> just for this one func) */
			for (len = 0; template[len]; ++len);
			
			loader = gdk_pixbuf_loader_new_with_type ("svg", &error);
			check_error (&error);
			
			gdk_pixbuf_loader_write (loader, (guchar *) template, len, &error);
			check_error (&error);
			
			gdk_pixbuf_loader_close(loader, &error);
			check_error (&error);

			pixbuf->array[i*2+j] = g_object_ref (gdk_pixbuf_loader_get_pixbuf(loader));

			g_object_unref (loader);
			g_free (template);
		}
	}

	/* free old pixbufs and the union */
	if (pixbuf_old) {
		for (i = 0; i < sizeof(pixbuf_old->array) / sizeof(pixbuf_old->array[0]); ++i)
			g_object_unref (pixbuf_old->array[i]);
		g_free (pixbuf_old);
	}

}

/* ------------------------------------------------------------------------- */

void save_config (void) {

	gchar		*filename;
	GKeyFile	*config;
	GError		*error = NULL;
	gsize		len;
	gchar		*contents;

	filename = g_build_filename (g_get_user_config_dir(), CONF_FILE, NULL);
	
	config = g_key_file_new();

	g_key_file_set_boolean (config, "Settings", "CapsLock",				show_caps);
	g_key_file_set_boolean (config, "Settings", "NumLock",				show_num);
	g_key_file_set_boolean (config, "Settings", "ScrollLock",			show_scroll);
	g_key_file_set_boolean (config, "Settings", "ReversedOrder",		rev_order);
	g_key_file_set_string (config,  "Settings", "Fg_Active",			fg_on);
	g_key_file_set_string (config,  "Settings", "Fg_Inactive",			fg_off);
	g_key_file_set_string (config,  "Settings", "Bg_Active_Beg",		bg_on_beg);
	g_key_file_set_string (config,  "Settings", "Bg_Active_End",		bg_on_end);
	g_key_file_set_string (config,  "Settings", "Bg_Inactive_Beg",		bg_off_beg);
	g_key_file_set_string (config,  "Settings", "Bg_Inactive_End",		bg_off_end);
	g_key_file_set_string (config,  "Settings", "CapsLock_Active",		sym_caps_on);
	g_key_file_set_string (config,  "Settings", "CapsLock_Inactive",	sym_caps_off);
	g_key_file_set_string (config,  "Settings", "NumLock_Active",		sym_num_on);
	g_key_file_set_string (config,  "Settings", "NumLock_Inactive",		sym_num_off);
	g_key_file_set_string (config,  "Settings", "ScrollLock_Active",	sym_scroll_on);
	g_key_file_set_string (config,  "Settings", "ScrollLock_Inactive",	sym_scroll_off);

	/* Grrr, just found out that older glib doesn't have this useful function! */
	//g_key_file_save_to_file (config, filename, &error);
	contents = g_key_file_to_data (config, &len, &error);
	check_error (&error);
	g_file_set_contents (filename, contents, len, &error);
	check_error (&error);
	
	g_key_file_free (config);
	g_free (filename);
	g_free (contents);
}

/* ------------------------------------------------------------------------- */

void helper_load_key (gint mode, GKeyFile *keyfile, gchar *group, gchar *key, gpointer target) {

	if (g_key_file_has_key (keyfile, group, key, NULL)) {
		switch (mode) {
			case 0:	/* gboolean */
				*(gboolean *) target = g_key_file_get_boolean (keyfile, group, key, NULL);
			break;
			case 1:	/* gchar * */
				g_strlcpy (target, g_key_file_get_string  (keyfile, group, key, NULL), -1);
			break;
		}
	}

}

/* ------------------------------------------------------------------------- */

void helper_regex_match (gchar *regex, gpointer target, gchar *defval) {

	if (!g_regex_match_simple (regex, target, 0, 0))
		g_strlcpy (target, defval, -1);
}

/* ------------------------------------------------------------------------- */

void load_config (void) {

	gchar		*filename;
	GKeyFile	*config;
	GError		*error = NULL;
	gchar *regex;

	filename = g_build_filename (g_get_user_config_dir(), CONF_FILE, NULL);
	
	if (!g_file_test (filename, G_FILE_TEST_EXISTS)) {
		g_free (filename);
		return;
	}

	config = g_key_file_new();
	g_key_file_load_from_file (config, filename, G_KEY_FILE_KEEP_COMMENTS, &error);

    if (error == NULL && g_key_file_has_group (config, "Settings")) {
	    helper_load_key (0, config, "Settings", "CapsLock",			&show_caps);
	    helper_load_key (0, config, "Settings", "NumLock",			&show_num);
	    helper_load_key (0, config, "Settings", "ScrollLock",		&show_scroll);
		helper_load_key (0, config, "Settings", "ReversedOrder",	&rev_order);
	    
		helper_load_key (1, config, "Settings", "Fg_Active",		fg_on);
		helper_load_key (1, config, "Settings", "Fg_Inactive",		fg_off);
		helper_load_key (1, config, "Settings", "Bg_Active_Beg", 	bg_on_beg);
		helper_load_key (1, config, "Settings", "Bg_Active_End", 	bg_on_end);
		helper_load_key (1, config, "Settings", "Bg_Inactive_Beg",	bg_off_beg);
		helper_load_key (1, config, "Settings", "Bg_Inactive_End",	bg_off_end);

		helper_load_key (1, config, "Settings", "CapsLock_Active",		sym_caps_on);
		helper_load_key (1, config, "Settings", "CapsLock_Inactive",	sym_caps_off);
		helper_load_key (1, config, "Settings", "NumLock_Active",		sym_num_on);
		helper_load_key (1, config, "Settings", "NumLock_Inactive",		sym_num_off);
		helper_load_key (1, config, "Settings", "ScrollLock_Active",	sym_scroll_on);
		helper_load_key (1, config, "Settings", "ScrollLock_Inactive",	sym_scroll_off);		
    } else {
        check_error (&error);
	}

	g_key_file_free (config);
	g_free (filename);

	/* sanity checks */	
	regex = "(?i)^#[0-9a-f]{6}$";
	helper_regex_match (regex, fg_on,		FG_ON);
	helper_regex_match (regex, fg_off,		FG_OFF);
	helper_regex_match (regex, bg_on_beg,	BG_ON_BEG);
	helper_regex_match (regex, bg_on_end,	BG_ON_END);
	helper_regex_match (regex, bg_off_beg,	BG_OFF_BEG);
	helper_regex_match (regex, bg_off_end,	BG_OFF_END);

	regex = "^.$";
	helper_regex_match (regex, sym_caps_on,		SYM_CAPS_ON);
	helper_regex_match (regex, sym_caps_off,	SYM_CAPS_OFF);
	helper_regex_match (regex, sym_num_on,		SYM_NUM_ON);
	helper_regex_match (regex, sym_num_off,		SYM_NUM_OFF);
	helper_regex_match (regex, sym_scroll_on,	SYM_SCROLL_ON);
	helper_regex_match (regex, sym_scroll_off,	SYM_SCROLL_OFF);

}

/* ------------------------------------------------------------------------- */

void toggle_state (GtkWidget *widget, gpointer data) {

	KeySym key;
	
	if ((GtkStatusIcon *) widget == icon_caps)
		key = XK_Caps_Lock;
	else if ((GtkStatusIcon *) widget == icon_num)
		key = XK_Num_Lock;
	else if ((GtkStatusIcon *) widget == icon_scroll)
		key = XK_Scroll_Lock;
	else
		return;

	
	Display* display = XOpenDisplay(NULL);
	XTestFakeKeyEvent (display, XKeysymToKeycode (display, key), True, CurrentTime);	/* press event */
	XTestFakeKeyEvent (display, XKeysymToKeycode (display, key), False, CurrentTime);	/* release event */
	XFlush (display);
	XCloseDisplay (display);

}

/* ------------------------------------------------------------------------- */	

void toggle_icon (GtkWidget *widget, gpointer *icon) {

	gboolean *state;

	if ((GtkStatusIcon *) *icon == icon_caps)
		state = &show_caps;
	else if ((GtkStatusIcon *) *icon == icon_num)
		state = &show_num;
	else if ((GtkStatusIcon *) *icon == icon_scroll)
		state = &show_scroll;
	else
		return;

	/* At least one icon has to be visible! */
	if ((show_caps + show_num + show_scroll <= 1) && *state == TRUE) {
		gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM(widget), TRUE);	/* Restore the tick */
		return;
	}
	
	//*state ^= 1;		/* toggle state */
	*state = !*state;	/* toggle state */
	
	gtk_status_icon_set_visible (*icon, *state);

	save_config();

}

/* ------------------------------------------------------------------------- */	

void update_icon (gint state, GtkStatusIcon *target, GdkPixbuf *icon_on, GdkPixbuf *icon_off, char *string) {
	
	gchar *tooltip;
	GdkPixbuf *icon;

	switch (state) {
		case 0:
			icon = icon_off;
			tooltip = g_strdup_printf ("%s %s", string, _("is OFF"));
			break;
		default:
			icon = icon_on;
			tooltip = g_strdup_printf ("%s %s", string, _("is ON"));
			break;
	}

	gtk_status_icon_set_from_pixbuf (target, icon);
	gtk_status_icon_set_tooltip_text (target, tooltip);

	g_free (tooltip);
}

/* ------------------------------------------------------------------------- */	

gint on_timeout (gpointer data) {

	static gint state_caps_old		= -1;
	static gint state_num_old		= -1;
	static gint state_scroll_old	= -1;
	
	if (refresh_request) {
		state_caps_old		= -1;
		state_num_old		= -1;
		state_scroll_old	= -1;
		refresh_request		= FALSE;
	}

	gint state_caps;
	gint state_num;
	gint state_scroll;	

	XKeyboardState state;

	Display *display = XOpenDisplay (NULL); 
	XGetKeyboardControl (display, &state);
	XCloseDisplay (display);

	state_caps		= state.led_mask & 1;
	state_num		= state.led_mask & 2;
	state_scroll	= state.led_mask & 4;
	
	if (state_caps != state_caps_old) {
		update_icon (state_caps, icon_caps, pixbuf->caps_on, pixbuf->caps_off, _("CapsLock"));
		state_caps_old = state_caps;
	}

	if (state_num != state_num_old) {
		update_icon (state_num, icon_num, pixbuf->num_on, pixbuf->num_off, _("NumLock"));		
		state_num_old = state_num;
	}

	if (state_scroll != state_scroll_old) {
		update_icon (state_scroll, icon_scroll, pixbuf->scroll_on, pixbuf->scroll_off, _("ScrollLock"));				
		state_scroll_old = state_scroll;
	}
	
	return TRUE;
}

/* ------------------------------------------------------------------------- */

void on_button_cancel (GtkWidget *widget, gpointer window) {
	gtk_widget_destroy (window);
}

/* ------------------------------------------------------------------------- */

void on_button_ok (GtkWidget *widget, gpointer window) {

	GdkColor gdk_col_tmp		= {0};
	
	GtkWidget *btn_fg_on		=	g_object_get_data (G_OBJECT(widget), "btn_fg_on");
	GtkWidget *btn_fg_off		=	g_object_get_data (G_OBJECT(widget), "btn_fg_off");
	GtkWidget *btn_bg_on_beg	=	g_object_get_data (G_OBJECT(widget), "btn_bg_on_beg");
	GtkWidget *btn_bg_on_end	=	g_object_get_data (G_OBJECT(widget), "btn_bg_on_end");
	GtkWidget *btn_bg_off_beg	=	g_object_get_data (G_OBJECT(widget), "btn_bg_off_beg");
	GtkWidget *btn_bg_off_end	=	g_object_get_data (G_OBJECT(widget), "btn_bg_off_end");
	GtkWidget *entry_caps_on	=	g_object_get_data (G_OBJECT(widget), "entry_caps_on");
	GtkWidget *entry_caps_off	=	g_object_get_data (G_OBJECT(widget), "entry_caps_off");
	GtkWidget *entry_num_on		=	g_object_get_data (G_OBJECT(widget), "entry_num_on");
	GtkWidget *entry_num_off	=	g_object_get_data (G_OBJECT(widget), "entry_num_off");
	GtkWidget *entry_scroll_on	=	g_object_get_data (G_OBJECT(widget), "entry_scroll_on");
	GtkWidget *entry_scroll_off	=	g_object_get_data (G_OBJECT(widget), "entry_scroll_off");
	GtkWidget *chk_order		=	g_object_get_data (G_OBJECT(widget), "chk_order");
	
	/* get colors and convert to #rrggbb */
	gtk_color_button_get_color (GTK_COLOR_BUTTON(btn_fg_on),		&gdk_col_tmp);
	GDK2HEX6 (fg_on, gdk_col_tmp);
	gtk_color_button_get_color (GTK_COLOR_BUTTON(btn_fg_off),		&gdk_col_tmp);
	GDK2HEX6 (fg_off, gdk_col_tmp);
	gtk_color_button_get_color (GTK_COLOR_BUTTON(btn_bg_on_beg),	&gdk_col_tmp);
	GDK2HEX6 (bg_on_beg, gdk_col_tmp);
	gtk_color_button_get_color (GTK_COLOR_BUTTON(btn_bg_on_end),	&gdk_col_tmp);
	GDK2HEX6 (bg_on_end, gdk_col_tmp);
	gtk_color_button_get_color (GTK_COLOR_BUTTON(btn_bg_off_beg),	&gdk_col_tmp);
	GDK2HEX6 (bg_off_beg, gdk_col_tmp);
	gtk_color_button_get_color (GTK_COLOR_BUTTON(btn_bg_off_end),	&gdk_col_tmp);
	GDK2HEX6 (bg_off_end, gdk_col_tmp);	

	/* get symbols */
	g_strlcpy (sym_caps_on,		gtk_entry_get_text (GTK_ENTRY(entry_caps_on)), -1);
	g_strlcpy (sym_caps_off,	gtk_entry_get_text (GTK_ENTRY(entry_caps_off)), -1);
	g_strlcpy (sym_num_on,		gtk_entry_get_text (GTK_ENTRY(entry_num_on)), -1);
	g_strlcpy (sym_num_off,		gtk_entry_get_text (GTK_ENTRY(entry_num_off)), -1);
	g_strlcpy (sym_scroll_on,	gtk_entry_get_text (GTK_ENTRY(entry_scroll_on)), -1);
	g_strlcpy (sym_scroll_off,	gtk_entry_get_text (GTK_ENTRY(entry_scroll_off)), -1);
	
	rev_order = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON(chk_order));
	
	gtk_widget_destroy (window);

	save_config();
	
	set_icons();
	refresh_request = TRUE;
}

/* ------------------------------------------------------------------------- */	

void on_refresh (GtkWidget *widget, gpointer data) {
	
	gtk_status_icon_set_visible (icon_caps,		FALSE);
	gtk_status_icon_set_visible (icon_num,		FALSE);
	gtk_status_icon_set_visible (icon_scroll,	FALSE);

	switch (rev_order) {
		case TRUE:
			gtk_status_icon_set_visible (icon_scroll, show_scroll);
			gtk_status_icon_set_visible (icon_num, show_num);
			gtk_status_icon_set_visible (icon_caps, show_caps);
			break;
		case FALSE:
			gtk_status_icon_set_visible (icon_caps, show_caps);
			gtk_status_icon_set_visible (icon_num, show_num);
			gtk_status_icon_set_visible (icon_scroll, show_scroll);
			break;
	}
}

/* ------------------------------------------------------------------------- */	

GtkWidget *helper_mk_table (GtkWidget *target, gchar *label, gint width, gint height) {

	GtkWidget *frame = gtk_frame_new (label);
	gtk_box_pack_start (GTK_BOX(target), frame, FALSE, FALSE, 0);

	GtkWidget *widget = gtk_table_new (width, height, FALSE);
	gtk_table_set_col_spacings (GTK_TABLE(widget), 8);
	gtk_table_set_row_spacings (GTK_TABLE(widget), 4);
	gtk_container_set_border_width (GTK_CONTAINER(widget), 4);
	gtk_container_add (GTK_CONTAINER(frame), widget);

	return widget;
}

/* ------------------------------------------------------------------------- */	

GtkWidget *helper_mk_label (GtkWidget *target, gchar *label, gfloat xalign, gfloat yalign, gint x1, gint x2, gint y1, gint y2) {

	GtkWidget *widget = gtk_label_new (label);
	gtk_misc_set_alignment (GTK_MISC(widget), xalign, yalign);
	gtk_table_attach_defaults (GTK_TABLE(target), widget, x1, x2, y1, y2);

	return widget;
}

/* ------------------------------------------------------------------------- */	

GtkWidget *helper_mk_entry (GtkWidget *target, gchar *symbol, gint x1, gint x2, gint y1, gint y2) {
	
	GtkWidget *widget = gtk_entry_new ();
	gtk_entry_set_text (GTK_ENTRY(widget), symbol);
	gtk_entry_set_max_length (GTK_ENTRY(widget), 1);
	gtk_entry_set_width_chars (GTK_ENTRY(widget), 1);
	gtk_entry_set_alignment (GTK_ENTRY(widget), 0.5);
	gtk_table_attach_defaults (GTK_TABLE(target), widget, x1, x2, y1, y2);
		
	return widget;
}
	
/* ------------------------------------------------------------------------- */	

GtkWidget *helper_mk_colorbutton (GtkWidget *target, gchar *color, gint x1, gint x2, gint y1, gint y2) {

	GdkColor gdk_col_tmp;

	gdk_color_parse (color, &gdk_col_tmp);
	GtkWidget *widget = gtk_color_button_new_with_color (&gdk_col_tmp);
	gtk_table_attach_defaults (GTK_TABLE(target), widget, x1, x2, y1, y2);
	
	return widget;
}

/* ------------------------------------------------------------------------- */	

void show_preferences (GtkWidget *widget, gpointer data) {
	
	GtkWidget *window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_window_set_title(GTK_WINDOW(window), _("Preferences"));
	gtk_window_set_resizable (GTK_WINDOW(window), FALSE);
	gtk_window_set_position (GTK_WINDOW(window), GTK_WIN_POS_MOUSE);
	gtk_window_set_icon_name (GTK_WINDOW(window), "gtk-preferences");
	
	GtkWidget *vbox = gtk_vbox_new (FALSE, 16);
	gtk_container_set_border_width (GTK_CONTAINER(vbox), 4);
	gtk_container_add (GTK_CONTAINER(window), vbox);

	GtkWidget *table_col = helper_mk_table (vbox, _("Colors"), 3, 6);
	
	helper_mk_label (table_col, _("Foreground (active):"), 0, 0.5, 0, 1, 0, 1);
	GtkWidget *btn_fg_on = helper_mk_colorbutton (table_col, fg_on, 1, 2, 0, 1);

	helper_mk_label (table_col, _("Foreground (inactive):"), 0, 0.5, 0, 1, 1, 2);
	GtkWidget *btn_fg_off = helper_mk_colorbutton (table_col, fg_off, 1, 2, 1, 2);
	helper_mk_label (table_col, _("Edge"), 0.5, 0.5, 1, 2, 2, 3);
	helper_mk_label (table_col, _("Center"), 0.5, 0.5, 3, 4, 2, 3);

	helper_mk_label (table_col, _("Background (active):"), 0, 0.5, 0, 1, 3, 4);
	GtkWidget *btn_bg_on_beg = helper_mk_colorbutton (table_col, bg_on_beg, 1, 2, 3, 4);	
	helper_mk_label (table_col, "->", 0.5, 0.5, 2, 3, 3, 4);
	GtkWidget *btn_bg_on_end = helper_mk_colorbutton (table_col, bg_on_end, 3, 4, 3, 4);

	helper_mk_label (table_col, _("Background (inactive):"), 0, 0.5, 0, 1, 4, 5);
	GtkWidget *btn_bg_off_beg = helper_mk_colorbutton (table_col, bg_off_beg, 1, 2, 4, 5);
	helper_mk_label (table_col, "->", 0.5, 0.5, 2, 3, 4, 5);
	GtkWidget *btn_bg_off_end = helper_mk_colorbutton (table_col, bg_off_end, 3, 4, 4, 5);


	GtkWidget *table_sym = helper_mk_table (vbox, _("Symbols"), 3, 4);
	
	helper_mk_label (table_sym, _("Active"), 		0.5, 0.5,	1, 2, 0, 1);
	helper_mk_label (table_sym, _("Inactive"), 		0.5, 0.5,	2, 3, 0, 1);
	helper_mk_label (table_sym, _("CapsLock:"), 	0, 0.5,		0, 1, 1, 2);
	helper_mk_label (table_sym, _("NumLock:"), 		0, 0.5,		0, 1, 2, 3);
	helper_mk_label (table_sym, _("ScrollLock:"),	0, 0.5,		0, 1, 3, 4);	

	GtkWidget *entry_caps_on	= helper_mk_entry (table_sym, sym_caps_on,		1, 2, 1, 2);
	GtkWidget *entry_caps_off	= helper_mk_entry (table_sym, sym_caps_off,		2, 3, 1, 2);
	GtkWidget *entry_num_on		= helper_mk_entry (table_sym, sym_num_on,		1, 2, 2, 3);
	GtkWidget *entry_num_off	= helper_mk_entry (table_sym, sym_num_off,		2, 3, 2, 3);
	GtkWidget *entry_scroll_on	= helper_mk_entry (table_sym, sym_scroll_on,	1, 2, 3, 4);
	GtkWidget *entry_scroll_off	= helper_mk_entry (table_sym, sym_scroll_off,	2, 3, 3, 4);
	

	GtkWidget *chk_align = gtk_alignment_new (0.5, 0.5, 0, 0);
	gtk_box_pack_start (GTK_BOX(vbox), chk_align, TRUE, TRUE, 0);

	GtkWidget *chk_order = gtk_check_button_new_with_label (_("Draw icons in reversed order"));
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(chk_order), rev_order);
	gtk_container_add (GTK_CONTAINER(chk_align), chk_order);

	GtkWidget *separator = gtk_hseparator_new();
	gtk_box_pack_start (GTK_BOX(vbox), separator, FALSE, FALSE, 0);

	GtkWidget *hbox = gtk_hbox_new (FALSE, 4);
	gtk_box_pack_start (GTK_BOX(vbox), hbox, FALSE, FALSE, 0);

	GtkWidget *btn_cancel = gtk_button_new_from_stock ("gtk-cancel");
	gtk_box_pack_end (GTK_BOX(hbox), btn_cancel, FALSE, FALSE, 0);
	g_signal_connect (G_OBJECT(btn_cancel), "clicked", G_CALLBACK(on_button_cancel), window);
	
	GtkWidget *btn_ok = gtk_button_new_from_stock ("gtk-ok");
	gtk_box_pack_end (GTK_BOX(hbox), btn_ok, FALSE, FALSE, 0);
	g_object_set_data (G_OBJECT(btn_ok), "btn_fg_on",			btn_fg_on);
	g_object_set_data (G_OBJECT(btn_ok), "btn_fg_off",			btn_fg_off);
	g_object_set_data (G_OBJECT(btn_ok), "btn_bg_on_beg",		btn_bg_on_beg);
	g_object_set_data (G_OBJECT(btn_ok), "btn_bg_on_end",		btn_bg_on_end);
	g_object_set_data (G_OBJECT(btn_ok), "btn_bg_off_beg",		btn_bg_off_beg);
	g_object_set_data (G_OBJECT(btn_ok), "btn_bg_off_end",		btn_bg_off_end);
	g_object_set_data (G_OBJECT(btn_ok), "entry_caps_on",		entry_caps_on);
	g_object_set_data (G_OBJECT(btn_ok), "entry_caps_off",		entry_caps_off);
	g_object_set_data (G_OBJECT(btn_ok), "entry_num_on",		entry_num_on);
	g_object_set_data (G_OBJECT(btn_ok), "entry_num_off",		entry_num_off);
	g_object_set_data (G_OBJECT(btn_ok), "entry_scroll_on",		entry_scroll_on);
	g_object_set_data (G_OBJECT(btn_ok), "entry_scroll_off",	entry_scroll_off);
	g_object_set_data (G_OBJECT(btn_ok), "chk_order",			chk_order);
	g_signal_connect (G_OBJECT(btn_ok), "clicked", G_CALLBACK(on_button_ok), window);
	
	gtk_widget_show_all (window);

}

/* ------------------------------------------------------------------------- */

void show_about (GtkWidget *widget, gpointer data) {

	gtk_window_set_default_icon_name ("gtk-about");
	gtk_show_about_dialog (NULL,
						   "program-name", APPNAME,
						   "copyright", CREDITS,
						   "website", WEBSITE,
						   "title", _("About"),
						   NULL);

}

/* ------------------------------------------------------------------------- */	

void show_popup_menu (GtkWidget *widget, guint button, guint time, gpointer menu) {

	gtk_menu_popup (GTK_MENU(menu), NULL, NULL, NULL, NULL, button, time);

}

/* ------------------------------------------------------------------------- */

GtkWidget *helper_mk_menuitem_check (gchar *label, GtkWidget *target, gpointer callback, gboolean state, GtkStatusIcon **icon) {

	GtkWidget *widget = gtk_check_menu_item_new_with_label (label);
	gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(widget), state);
	gtk_menu_shell_append (GTK_MENU_SHELL (target), widget);
	g_signal_connect (G_OBJECT (widget), "toggled", G_CALLBACK(callback), (gpointer) icon);

	return widget;
}

/* ------------------------------------------------------------------------- */

GtkWidget *helper_mk_menuitem_stock (gchar *stock, gchar *label, GtkWidget *target, gpointer callback) {

	GtkWidget *widget = gtk_image_menu_item_new_from_stock (stock, NULL);
	if (label) gtk_menu_item_set_label (GTK_MENU_ITEM(widget), label);
	gtk_menu_shell_append (GTK_MENU_SHELL (target), widget);
	g_signal_connect (G_OBJECT (widget), "activate", G_CALLBACK(callback), NULL);

	return widget;
}

/* ------------------------------------------------------------------------- */

GtkStatusIcon *helper_mk_status_icon (gpointer callback_left, gpointer callback_right, GtkWidget *menu) {

	GtkStatusIcon *widget = gtk_status_icon_new ();
	gtk_status_icon_set_visible (widget, FALSE);
	g_signal_connect (G_OBJECT (widget), "activate", G_CALLBACK (callback_left), NULL);
	g_signal_connect (G_OBJECT (widget), "popup-menu", G_CALLBACK (callback_right), menu);

	return widget;
}

/* ------------------------------------------------------------------------- */

void make_tray_icons () {

	GtkWidget *menu = gtk_menu_new();

	helper_mk_menuitem_check (_("CapsLock"),	menu, toggle_icon, show_caps,	&icon_caps);
	helper_mk_menuitem_check (_("NumLock"),		menu, toggle_icon, show_num,	&icon_num);
	helper_mk_menuitem_check (_("ScrollLock"),	menu, toggle_icon, show_scroll,	&icon_scroll);

	GtkWidget *separator = gtk_separator_menu_item_new();
	gtk_menu_shell_append (GTK_MENU_SHELL (menu), separator);

    helper_mk_menuitem_stock ("gtk-refresh",		_("Redraw icons"), menu, on_refresh);    
    helper_mk_menuitem_stock ("gtk-preferences",	NULL, menu, show_preferences);
    helper_mk_menuitem_stock ("gtk-about",			NULL, menu, show_about);
    helper_mk_menuitem_stock ("gtk-quit",			NULL, menu, gtk_main_quit);

	icon_caps	= helper_mk_status_icon (toggle_state, show_popup_menu, menu);
	icon_num	= helper_mk_status_icon (toggle_state, show_popup_menu, menu);
	icon_scroll	= helper_mk_status_icon (toggle_state, show_popup_menu, menu);
	
	gtk_timeout_add (100, on_timeout, NULL);

	gtk_widget_show_all (menu);

	on_refresh(NULL, NULL);

}

/* ========================================================================= */

gint main(gint argc, gchar *argv[]) {

	/* Gettext */
	setlocale (LC_ALL, "");
	bindtextdomain ("lockstate", "/usr/share/locale/");
	textdomain ("lockstate");
	bind_textdomain_codeset ("lockstate", "UTF-8");

	/* Init */
	gtk_init (&argc, &argv);
	load_config();
	set_icons();

	/* Precaution */
	if (!show_caps && !show_num && !show_scroll) show_caps = TRUE;
	
	make_tray_icons ();
	gtk_main();
	
	return 0;
}

/* ========================================================================= */	
